@@ -5,10 +5,10 @@ This file shamelessly taken from the requests library"""
 __title__ = 'simple-salesforce-propic'
 __description__ = 'A basic Salesforce.com REST API client.'
 __url__ = 'https://github.com/simple-salesforce/simple-salesforce'
-__version__ = '0.74.4'
+__version__ = '1.11.0'
 __author__ = 'Nick Catalano'
 __author_email__ = 'nickcatal@gmail.com'
 __license__ = 'Apache 2.0'
-__maintainer__ = 'Demian Brecht'
-__maintainer_email__ = 'demianbrecht@gmail.com'
+__maintainer__ = 'Jonathan Wobken'
+__maintainer_email__ = 'jonathanwobken@gmail.com'
 __keywords__ = 'python salesforce salesforce.com'
