@@ -1,66 +1,55 @@
 """Core classes and exceptions for Simple-Salesforce"""
 
-
 # has to be defined prior to login import
-DEFAULT_API_VERSION = '38.0'
+DEFAULT_API_VERSION = '42.0'
 
-
-import logging
-import warnings
-import requests
 import json
+import logging
 import re
-from collections import namedtuple
+from collections import OrderedDict, namedtuple
+from urllib.parse import urljoin, urlparse
 
-try:
-    from urlparse import urlparse, urljoin
-except ImportError:
-    # Python 3+
-    from urllib.parse import urlparse, urljoin
+import requests
 
-from simple_salesforce.login import SalesforceLogin
-from simple_salesforce.util import date_to_iso8601, exception_handler
-from simple_salesforce.exceptions import (
-    SalesforceGeneralError
-)
-from simple_salesforce.bulk import SFBulkHandler
+from .bulk import SFBulkHandler
+from .exceptions import SalesforceGeneralError
+from .login import SalesforceLogin
+from .util import date_to_iso8601, exception_handler
+from .metadata import SfdcMetadataApi
 
-try:
-    from collections import OrderedDict
-except ImportError:
-    # Python < 2.7
-    from ordereddict import OrderedDict
-
-#pylint: disable=invalid-name
+# pylint: disable=invalid-name
 logger = logging.getLogger(__name__)
 
-
-def _warn_request_deprecation():
-    """Deprecation for (Salesforce/SFType).request attribute"""
-    warnings.warn(
-        'The request attribute has been deprecated and will be removed in a '
-        'future version. Please use Salesforce.session instead.',
-        DeprecationWarning
-    )
-
-
 Usage = namedtuple('Usage', 'used total')
 PerAppUsage = namedtuple('PerAppUsage', 'used total name')
 
 
 # pylint: disable=too-many-instance-attributes
-class Salesforce(object):
+class Salesforce:
     """Salesforce Instance
 
     An instance of Salesforce is a handy way to wrap a Salesforce session
     for easy use of the Salesforce REST API.
     """
-    # pylint: disable=too-many-arguments
+
+    # pylint: disable=too-many-arguments,too-many-locals,too-many-branches
     def __init__(
-            self, username=None, password=None, security_token=None,
-            session_id=None, instance=None, instance_url=None,
-            organizationId=None, sandbox=None, version=DEFAULT_API_VERSION,
-            proxies=None, session=None, client_id=None, domain=None):
+            self,
+            username=None,
+            password=None,
+            security_token=None,
+            session_id=None,
+            instance=None,
+            instance_url=None,
+            organizationId=None,
+            version=DEFAULT_API_VERSION,
+            proxies=None,
+            session=None,
+            client_id=None,
+            domain=None,
+            consumer_key=None,
+            privatekey_file=None,
+            ):
         """Initialize the instance with the given parameters.
 
         Available kwargs
@@ -70,12 +59,17 @@ class Salesforce(object):
         * username -- the Salesforce username to use for authentication
         * password -- the password for the username
         * security_token -- the security token for the username
-        * sandbox -- DEPRECATED: Use domain instead.
         * domain -- The domain to using for connecting to Salesforce. Use
                     common domains, such as 'login' or 'test', or
                     Salesforce My domain. If not used, will default to
                     'login'.
 
+        OAuth 2.0 JWT Bearer Token Authentication:
+
+        * consumer_key -- the consumer key generated for the user
+        * privatekey_file -- the path to the private key file used
+                             for signing the JWT token
+
         Direct Session and Instance Access:
 
         * session_id -- Access token for this session
@@ -96,18 +90,6 @@ class Salesforce(object):
                      exposed by simple_salesforce.
 
         """
-        if (sandbox is not None) and (domain is not None):
-            raise ValueError("Both 'sandbox' and 'domain' arguments were "
-                             "supplied. Either may be supplied, but not "
-                             "both.")
-
-        if sandbox is not None:
-            warnings.warn("'sandbox' argument is deprecated. Use "
-                          "'domain' instead. Overriding 'domain' "
-                          "with 'sandbox' value.",
-                          DeprecationWarning)
-
-            domain = 'test' if sandbox else 'login'
 
         if domain is None:
             domain = 'login'
@@ -126,7 +108,7 @@ class Salesforce(object):
                 logger.warning(
                     'Proxies must be defined on custom session object, '
                     'ignoring proxies: %s', proxies
-                )
+                    )
 
         # Determine if the user wants to use our username/password auth or pass
         # in their own information
@@ -172,10 +154,23 @@ class Salesforce(object):
                 client_id=client_id,
                 domain=self.domain)
 
+        elif all(arg is not None for arg in (
+                username, consumer_key, privatekey_file)):
+            self.auth_type = "jwt-bearer"
+
+            # Pass along the username/password to our login helper
+            self.session_id, self.sf_instance = SalesforceLogin(
+                session=self.session,
+                username=username,
+                consumer_key=consumer_key,
+                privatekey_file=privatekey_file,
+                proxies=self.proxies,
+                domain=self.domain)
+
         else:
             raise TypeError(
                 'You must provide login information or an instance and token'
-            )
+                )
 
         self.auth_site = ('https://{domain}.salesforce.com'
                           .format(domain=self.domain))
@@ -184,7 +179,7 @@ class Salesforce(object):
             'Content-Type': 'application/json',
             'Authorization': 'Bearer ' + self.session_id,
             'X-PrettyPrint': '1'
-        }
+            }
 
         self.base_url = ('https://{instance}/services/data/v{version}/'
                          .format(instance=self.sf_instance,
@@ -194,14 +189,21 @@ class Salesforce(object):
         self.bulk_url = ('https://{instance}/services/async/{version}/'
                          .format(instance=self.sf_instance,
                                  version=self.sf_version))
+        self.metadata_url = ('https://{instance}/services/Soap/m/{version}/'
+                             .format(instance=self.sf_instance,
+                                     version=self.sf_version))
 
         self.api_usage = {}
 
-    def describe(self):
+    def describe(self, **kwargs):
         """Describes all available objects
+
+        Arguments:
+
+        * keyword arguments supported by requests.request (e.g. json, timeout)
         """
         url = self.base_url + "sobjects"
-        result = self._call_salesforce('GET', url, name='describe')
+        result = self._call_salesforce('GET', url, name='describe', **kwargs)
 
         json_result = result.json(object_pairs_hook=OrderedDict)
         if len(json_result) == 0:
@@ -209,6 +211,17 @@ class Salesforce(object):
 
         return json_result
 
+    def is_sandbox(self):
+        """After connection returns is the organization is a sandbox"""
+        is_sandbox = None
+        if self.session_id:
+            is_sandbox = self.query_all("SELECT IsSandbox "
+                                        "FROM Organization LIMIT 1")
+            is_sandbox = is_sandbox.get('records', [{'IsSandbox': None}])[
+                0].get(
+                'IsSandbox')
+        return is_sandbox
+
     # SObject Handler
     def __getattr__(self, name):
         """Returns an `SFType` instance for the given Salesforce object type
@@ -227,7 +240,7 @@ class Salesforce(object):
         # fix to enable serialization
         # (https://github.com/heroku/simple-salesforce/issues/60)
         if name.startswith('__'):
-            return super(Salesforce, self).__getattr__(name)
+            return super().__getattr__(name)
 
         if name == 'bulk':
             # Deal with bulk API functions
@@ -243,7 +256,8 @@ class Salesforce(object):
         """Sets the password of a user
 
         salesforce dev documentation link:
-        https://www.salesforce.com/us/developer/docs/api_rest/Content/dome_sobject_user_password.htm
+        https://www.salesforce.com/us/developer/docs/api_rest/Content
+        /dome_sobject_user_password.htm
 
         Arguments:
 
@@ -268,25 +282,6 @@ class Salesforce(object):
 
         return json_result
 
-    # pylint: disable=invalid-name
-    def setPassword(self, user, password):
-        # pylint: disable=line-too-long
-        """Sets the password of a user
-
-        salesforce dev documentation link:
-        https://www.salesforce.com/us/developer/docs/api_rest/Content/dome_sobject_user_password.htm
-
-        Arguments:
-
-        * user: the userID of the user to set
-        * password: the new password
-        """
-        warnings.warn(
-            "This method has been deprecated."
-            "Please use set_password instead.",
-            DeprecationWarning)
-        return self.set_password(user, password)
-
     # Generic Rest Function
     def restful(self, path, params=None, method='GET', **kwargs):
         """Allows you to make a direct REST call if you know the path
@@ -342,7 +337,7 @@ class Salesforce(object):
                     string will be wrapped to read `FIND {Waldo}` before being
                     sent to Salesforce
         """
-        search_string = u'FIND {{{search_string}}}'.format(search_string=search)
+        search_string = 'FIND {{{search_string}}}'.format(search_string=search)
         return self.search(search_string)
 
     def limits(self, **kwargs):
@@ -398,7 +393,7 @@ class Salesforce(object):
         """
         if identifier_is_url:
             # Don't use `self.base_url` here because the full URI is provided
-            url = (u'https://{instance}{next_record_url}'
+            url = ('https://{instance}{next_record_url}'
                    .format(instance=self.sf_instance,
                            next_record_url=next_records_identifier))
         else:
@@ -410,6 +405,37 @@ class Salesforce(object):
 
         return result.json(object_pairs_hook=OrderedDict)
 
+    def query_all_iter(self, query, include_deleted=False, **kwargs):
+        """This is a lazy alternative to `query_all` - it does not construct
+        the whole result set into one container, but returns objects from each
+        page it retrieves from the API.
+
+        Since `query_all` has always been eagerly executed, we reimplemented it
+        using `query_all_iter`, only materializing the returned iterator to
+        maintain backwards compatibility.
+
+        The one big difference from `query_all` (apart from being lazy) is that
+        we don't return a dictionary with `totalSize` and `done` here,
+        we only return the records in an iterator.
+
+        Arguments
+
+        * query -- the SOQL query to send to Salesforce, e.g.
+                   SELECT Id FROM Lead WHERE Email = "waldo@somewhere.com"
+        * include_deleted -- True if the query should include deleted records.
+        """
+
+        result = self.query(query, include_deleted=include_deleted, **kwargs)
+        while True:
+            for record in result['records']:
+                yield record
+            # fetch next batch if we're not done else break out of loop
+            if not result['done']:
+                result = self.query_more(result['nextRecordsUrl'],
+                                         identifier_is_url=True)
+            else:
+                return
+
     def query_all(self, query, include_deleted=False, **kwargs):
         """Returns the full set of results for the `query`. This is a
         convenience
@@ -427,20 +453,14 @@ class Salesforce(object):
         * include_deleted -- True if the query should include deleted records.
         """
 
-        result = self.query(query, include_deleted=include_deleted, **kwargs)
-        all_records = []
-
-        while True:
-            all_records.extend(result['records'])
-            # fetch next batch if we're not done else break out of loop
-            if not result['done']:
-                result = self.query_more(result['nextRecordsUrl'],
-                                         identifier_is_url=True)
-            else:
-                break
-
-        result['records'] = all_records
-        return result
+        records = self.query_all_iter(query, include_deleted=include_deleted,
+                                      **kwargs)
+        all_records = list(records)
+        return {
+            'records': all_records,
+            'totalSize': len(all_records),
+            'done': True,
+            }
 
     def apexecute(self, action, method='GET', data=None, **kwargs):
         """Makes an HTTP request to an APEX REST endpoint
@@ -452,12 +472,15 @@ class Salesforce(object):
         * data -- A dict of parameters to send in a POST / PUT request
         * kwargs -- Additional kwargs to pass to `requests.request`
         """
+        # If data is None, we should send an empty body, not "null", which is
+        # None in json.
+        json_data = json.dumps(data) if data is not None else None
         result = self._call_salesforce(
             method,
             self.apex_url + action,
             name="apexexcute",
-            data=json.dumps(data), **kwargs
-        )
+            data=json_data, **kwargs
+            )
         try:
             response_content = result.json()
         # pylint: disable=broad-except
@@ -487,18 +510,6 @@ class Salesforce(object):
 
         return result
 
-    @property
-    def request(self):
-        """Deprecated access to self.session for backwards compatibility"""
-        _warn_request_deprecation()
-        return self.session
-
-    @request.setter
-    def request(self, session):
-        """Deprecated setter for self.session"""
-        _warn_request_deprecation()
-        self.session = session
-
     @staticmethod
     def parse_api_usage(sforce_limit_info):
         """parse API usage and limits out of the Sforce-Limit-Info header
@@ -529,13 +540,78 @@ class Salesforce(object):
 
         return result
 
-class SFType(object):
+    # file-based deployment function
+    def deploy(self, zipfile, sandbox, **kwargs):
+
+        """Deploy using the Salesforce Metadata API. Wrapper for
+        SfdcMetaDataApi.deploy(...).
+
+        Arguments:
+
+        * zipfile: a .zip archive to deploy to an org, given as (
+        "path/to/zipfile.zip")
+        * options: salesforce DeployOptions in .json format.
+            (https://developer.salesforce.com/docs/atlas.en-us.api_meta.meta
+            /api_meta/meta_deploy.htm)
+
+        Returns a process id and state for this deployment.
+        """
+        mdapi = SfdcMetadataApi(session=self.session,
+                                session_id=self.session_id,
+                                instance=self.sf_instance,
+                                sandbox=sandbox,
+                                metadata_url=self.metadata_url,
+                                api_version=self.sf_version,
+                                headers=self.headers)
+        asyncId, state = mdapi.deploy(zipfile, **kwargs)
+        result = {'asyncId': asyncId, 'state': state}
+        return result
+
+    # check on a file-based deployment
+    def checkDeployStatus(self, asyncId, sandbox, **kwargs):
+        """Check on the progress of a file-based deployment via Salesforce
+        Metadata API.
+        Wrapper for SfdcMetaDataApi.check_deploy_status(...).
+
+        Arguments:
+
+        * asyncId: deployment async process ID, returned by Salesforce.deploy()
+
+        Returns status of the deployment the asyncId given.
+        """
+        mdapi = SfdcMetadataApi(session=self.session,
+                                session_id=self.session_id,
+                                instance=self.sf_instance,
+                                sandbox=sandbox,
+                                metadata_url=self.metadata_url,
+                                api_version=self.sf_version,
+                                headers=self.headers)
+
+        state, state_detail, deployment_detail, unit_test_detail = \
+            mdapi.check_deploy_status(asyncId, **kwargs)
+        results = {
+            'state': state,
+            'state_detail': state_detail,
+            'deployment_detail': deployment_detail,
+            'unit_test_detail': unit_test_detail
+            }
+
+        return results
+
+
+class SFType:
     """An interface to a specific type of SObject"""
 
     # pylint: disable=too-many-arguments
     def __init__(
-            self, object_name, session_id, sf_instance,
-            sf_version=DEFAULT_API_VERSION, proxies=None, session=None):
+            self,
+            object_name,
+            session_id,
+            sf_instance,
+            sf_version=DEFAULT_API_VERSION,
+            proxies=None,
+            session=None,
+            ):
         """Initialize the instance with the given parameters.
 
         Arguments:
@@ -559,7 +635,7 @@ class SFType(object):
         self.api_usage = {}
 
         self.base_url = (
-            u'https://{instance}/services/data/v{sf_version}/sobjects'
+            'https://{instance}/services/data/v{sf_version}/sobjects'
             '/{object_name}/'.format(instance=sf_instance,
                                      object_name=object_name,
                                      sf_version=sf_version))
@@ -586,7 +662,7 @@ class SFType(object):
         result = self._call_salesforce(
             method='GET', url=urljoin(self.base_url, 'describe'),
             headers=headers
-        )
+            )
         return result.json(object_pairs_hook=OrderedDict)
 
     def describe_layout(self, record_id, headers=None):
@@ -603,12 +679,12 @@ class SFType(object):
         """
         custom_url_part = 'describe/layouts/{record_id}'.format(
             record_id=record_id
-        )
+            )
         result = self._call_salesforce(
             method='GET',
             url=urljoin(self.base_url, custom_url_part),
             headers=headers
-        )
+            )
         return result.json(object_pairs_hook=OrderedDict)
 
     def get(self, record_id, headers=None):
@@ -623,7 +699,7 @@ class SFType(object):
         result = self._call_salesforce(
             method='GET', url=urljoin(self.base_url, record_id),
             headers=headers
-        )
+            )
         return result.json(object_pairs_hook=OrderedDict)
 
     def get_by_custom_id(self, custom_id_field, custom_id, headers=None):
@@ -643,11 +719,11 @@ class SFType(object):
         custom_url = urljoin(
             self.base_url, '{custom_id_field}/{custom_id}'.format(
                 custom_id_field=custom_id_field, custom_id=custom_id
+                )
             )
-        )
         result = self._call_salesforce(
             method='GET', url=custom_url, headers=headers
-        )
+            )
         return result.json(object_pairs_hook=OrderedDict)
 
     def create(self, data, headers=None):
@@ -664,7 +740,7 @@ class SFType(object):
         result = self._call_salesforce(
             method='POST', url=self.base_url,
             data=json.dumps(data), headers=headers
-        )
+            )
         return result.json(object_pairs_hook=OrderedDict)
 
     def upsert(self, record_id, data, raw_response=False, headers=None):
@@ -688,7 +764,7 @@ class SFType(object):
         result = self._call_salesforce(
             method='PATCH', url=urljoin(self.base_url, record_id),
             data=json.dumps(data), headers=headers
-        )
+            )
         return self._raw_response(result, raw_response)
 
     def update(self, record_id, data, raw_response=False, headers=None):
@@ -711,7 +787,7 @@ class SFType(object):
         result = self._call_salesforce(
             method='PATCH', url=urljoin(self.base_url, record_id),
             data=json.dumps(data), headers=headers
-        )
+            )
         return self._raw_response(result, raw_response)
 
     def delete(self, record_id, raw_response=False, headers=None):
@@ -732,7 +808,7 @@ class SFType(object):
         result = self._call_salesforce(
             method='DELETE', url=urljoin(self.base_url, record_id),
             headers=headers
-        )
+            )
         return self._raw_response(result, raw_response)
 
     def deleted(self, start, end, headers=None):
@@ -741,7 +817,8 @@ class SFType(object):
 
         Use the SObject Get Deleted resource to get a list of deleted records
         for the specified object.
-        .../deleted/?start=2013-05-05T00:00:00+00:00&end=2013-05-10T00:00:00+00:00
+        .../deleted/?start=2013-05-05T00:00:00+00:00&end=2013-05-10T00:00:00
+        +00:00
 
         * start -- start datetime object
         * end -- end datetime object
@@ -750,8 +827,8 @@ class SFType(object):
         url = urljoin(
             self.base_url, 'deleted/?start={start}&end={end}'.format(
                 start=date_to_iso8601(start), end=date_to_iso8601(end)
+                )
             )
-        )
         result = self._call_salesforce(method='GET', url=url, headers=headers)
         return result.json(object_pairs_hook=OrderedDict)
 
@@ -762,7 +839,8 @@ class SFType(object):
         Use the SObject Get Updated resource to get a list of updated
         (modified or added) records for the specified object.
 
-         .../updated/?start=2014-03-20T00:00:00+00:00&end=2014-03-22T00:00:00+00:00
+         .../updated/?start=2014-03-20T00:00:00+00:00&end=2014-03-22T00:00:00
+         +00:00
 
         * start -- start datetime object
         * end -- end datetime object
@@ -771,8 +849,8 @@ class SFType(object):
         url = urljoin(
             self.base_url, 'updated/?start={start}&end={end}'.format(
                 start=date_to_iso8601(start), end=date_to_iso8601(end)
+                )
             )
-        )
         result = self._call_salesforce(method='GET', url=url, headers=headers)
         return result.json(object_pairs_hook=OrderedDict)
 
@@ -785,7 +863,7 @@ class SFType(object):
             'Content-Type': 'application/json',
             'Authorization': 'Bearer ' + self.session_id,
             'X-PrettyPrint': '1'
-        }
+            }
         additional_headers = kwargs.pop('headers', dict())
         headers.update(additional_headers or dict())
         result = self.session.request(method, url, headers=headers, **kwargs)
@@ -810,50 +888,3 @@ class SFType(object):
             return response.status_code
 
         return response
-
-    @property
-    def request(self):
-        """Deprecated access to self.session for backwards compatibility"""
-        _warn_request_deprecation()
-        return self.session
-
-    @request.setter
-    def request(self, session):
-        """Deprecated setter for self.session"""
-        _warn_request_deprecation()
-        self.session = session
-
-
-class SalesforceAPI(Salesforce):
-    """Deprecated SalesforceAPI Instance
-
-    This class implements the Username/Password Authentication Mechanism using
-    Arguments It has since been surpassed by the 'Salesforce' class, which
-    relies on kwargs
-
-    """
-    # pylint: disable=too-many-arguments
-    def __init__(self, username, password, security_token, sandbox=False,
-                 sf_version='27.0'):
-        """Initialize the instance with the given parameters.
-
-        Arguments:
-
-        * username -- the Salesforce username to use for authentication
-        * password -- the password for the username
-        * security_token -- the security token for the username
-        * sandbox -- True if you want to login to `test.salesforce.com`, False
-                     if you want to login to `login.salesforce.com`.
-        * sf_version -- the version of the Salesforce API to use, for example
-                        "27.0"
-        """
-        warnings.warn(
-            "Use of login arguments has been deprecated. Please use kwargs",
-            DeprecationWarning
-        )
-
-        super(SalesforceAPI, self).__init__(username=username,
-                                            password=password,
-                                            security_token=security_token,
-                                            sandbox=sandbox,
-                                            version=sf_version)
